/*****************************************************************************/
/*
                                Logging.c

Logging functions for the WASD VMS HTTPd. By default, produces a Web "common"
format log suitable for analysis by tools capable of processing that format.
Secretes four pseudo-requests logging the BEGINing, OPENing, CLOSEing and
ENDing of logging activities. These shouldn't introduce problems for analysis
tools as they look like legitimate POST entries for user "HHTPd"!

Also can produce logs in the "common+server", "combined" and a user-defined
format.  If a user-defined format contains an error the server will exit the
first time a request is logged.  Caveat emptor!

Log file name must be supplied via 'LoggingFileName' or the logical name 
HTTPD$LOG (process, job, or system level).  An error is reporting if logging 
is enabled without either of these being supplied.  If the log file name
specified is "SYS$OUTPUT" log entries are written to SYS$OUTPUT.  This is
useful for checking user-defined log entries before committing files to that
format.


COMMON FORMAT
-------------
(NCSA, CERN, etc.)

'client rfc891 auth-user date/time request status bytes'


COMBINED FORMAT
---------------
(NCSA)

'client rfc891 auth-user date/time request status bytes referer user-agent'


COMMON FORMAT plus server name
-------------
(NCSA, CERN, etc.)

'client rfc891 auth-user date/time request status bytes server'


USER-DEFINED FORMAT
-------------------

Must begin with a character that is used as a substitute when a particular
field is empty (use "\0" for no substitute, as in the "windows log format"
example below).  Two different "escape" characters introduce the following
parameters of characters.

A '!' followed by ...

  'AR' :  authentication realm
  'AU' :  authenticated user name
  'BB' :  bytes in body (excluding response header)
  'BY' :  bytes transmitted to the client
  'CN' :  client host name (or address if DNS lookup disabled)
  'CA' :  client numeric address
  'EM' :  request elapsed time in milliseconds
  'ES' :  request elapsed time in seconds
  'ME' :  request method
  'PA' :  request path (NOTE ... this is not the full request path!  See 'R')
  'QS' :  request query string
  'RF' :  referer
  'RQ' :  FULL request path (including script and any query string)
  'RS' :  response status code
  'SC' :  script name
  'SN' :  server host name
  'SP' :  server port
  'TC' :  request time (common log format)
  'TG' :  request time (GMT)
  'TV' :  request time (VMS format)
  'UA' :  user agent

A '\' followed by ...

  '0' :  a null character (used to define the empty field character)
  '!' :  insert an '!'
  '\' :  insert a '\'
  'n' :  insert a newline
  'q' :  insert a quote (so that in the DCL the quotes won't need escaping!)
  't' :  insert a TAB

Any other character is directly inserted into the log entry.


Examples
--------

The equivalent of the common log format is:

  "-!CN - !AU [!TC] \q!RQ\q !RS !BY"

The combined log format:

  "-!CN - !AU [!TC] \q!RQ\q !RS !BY \q!RF\q \q!UA\q"

The WebSite "windows log format" would be created by:

  "\0!TC\t!CA\t!SN\t!AR\t!AU\t!ME\t!PA\t!RQ\t!EM\t!UA\t!RS\t!BB\t"

To get the common log format plus append request duration in seconds:

  "-!CN - !AU [!TC] \q!RQ\q !RS !BY !ES"


VERSION HISTORY
---------------
12-OCT-97  MGD  bugfix; user format from configuration file
25-SEP-97  MGD  bugfix; ACCVIO if request structure exists but empty
                (assumed if it existed all the pointers were valid - WRONG!)
06-SEP-97  MGD  added "common+server", "combined", "user-defined" log formats
10-JUL-97  MGD  minor modification for server-admin logging on/off/flush
10-JUL-96  MGD  ResourcePtr instead of ScriptPtr/PathInfoPtr/QueryStringPtr
01-DEC-95  MGD  HTTPd version 3, "common" log format
27-SEP-95  MGD  use time values in request thread structure
16-JUN-95  MGD  added node name, changed logging fields to be space-separated
07-APR-95  MGD  initial development for addition to multi-threaded daemon
*/
/*****************************************************************************/

/* standard C header files */
#include <stdio.h>
#include <ctype.h>
#include <errno.h>

/* VMS related header files */
#include <descrip.h>
#include <iodef.h>
#include <lnmdef.h>
#include <rms.h>
#include <ssdef.h>
#include <stsdef.h>

/* application header files */
#include "wasd.h"
#include "httpd.h"
#include "error.h"
#include "logging.h"

/******************/
/* global storage */
/******************/

char  ErrorLoggingFormatLength [] = "log format too long";

boolean  LoggingEnabled = false,
         LoggingFormatCommon = false,
         LoggingFormatCommonServer = false,
         LoggingFormatCombined = false,
         LoggingFileOpen = false,
         LoggingFlushTimerOutstanding = false,
         LoggingSysOutput = false;

char  LoggingFileName [256],
      LoggingFormatUser [256];

/********************/
/* external storage */
/********************/

#ifdef DBUG
extern boolean Debug;
#else
#define Debug 0 
#endif

extern int  ExitStatus;
extern char  ErrorSanityCheck[];
extern char  ServerHostName[];
extern char  SoftwareID[];
extern char  TimeGmtString[];
extern char  Utility[];
extern struct ConfigStruct  Config;

/*****************************************************************************/
/*
Web "common" format log.

Parameter 'Function' can be:

LOGGING_BEGIN           server startup
LOGGING_OPEN            open server logging, other than server startup
LOGGING_CLOSE           close server logging, other than server shutdown
LOGGING_END             server shutdown
LOGGING_ENTRY           log a request
LOGGING_FLUSH           close the log file, flushing the contents
*/ 

int Logging
(
struct RequestStruct *rqptr,
int Function
)
{
#  define MAX_FAO_VECTOR 64

   static char  *MonthName [] =
      { "", "Jan", "Feb", "Mar", "Apr", "May", "Jun",
            "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

   /* flush the log file one minute after the first entry after flushing */
   static $DESCRIPTOR (FlushDeltaTimeDsc, "0 00:01:00.00");

   static $DESCRIPTOR (BytesFaoDsc, "!UL\0");
   static $DESCRIPTOR (ElapsedResponseDurationDsc, "!UL\0");
   static $DESCRIPTOR (ElapsedSecondsDsc, "!UL.!3ZL\0");
   static $DESCRIPTOR (StatusFaoDsc, "!UL\0");
   static $DESCRIPTOR (VmsTimeFaoDsc, "!%D\0");

   static $DESCRIPTOR (LogFileTimeFaoDsc,
                       "!2ZL/!3AZ/!4ZL:!2ZL:!2ZL:!2ZL !AZ");

   /* server-host - - [date/time] \"what status\" 200 0 */
   static $DESCRIPTOR (LogFileBogusFaoDsc,
                       "!AZ - HTTPd [!AZ] \"POST !AZ-!8XL\" 200 -!AZ");

   /* host r-ident user [date/time] \"request\" status bytes */
   static $DESCRIPTOR (LogFileCommonFaoDsc,
                       "!AZ - !AZ [!AZ] \"!AZ !AZ\" !UL !AZ");

   /* as for "common" above, plus ' server'*/
   static $DESCRIPTOR (LogFileCommonServerFaoDsc,
                       "!AZ - !AZ [!AZ] \"!AZ !AZ\" !UL !AZ !AZ");

   /* as for "common" above, plus ' referer user-agent'*/
   static $DESCRIPTOR (LogFileCombinedFaoDsc,
                       "!AZ - !AZ [!AZ] \"!AZ !AZ\" !UL !AZ \"!AZ\" \"!AZ\"");

   /* user-defined log-file format */
   static $DESCRIPTOR (LogFileUserFaoDsc, "!#(AZ)");

   static $DESCRIPTOR (LnmFileDevDsc, "LNM$FILE_DEV");
   static $DESCRIPTOR (LoggingFileNameLogicalDsc, "HTTPD$LOG");

   static boolean  UserFormatProblem = false;

   static int  LoggingFileNameLength = 0,
               RecordCount = 0;
   static unsigned short  Length,
                          NumericTime [7];
   static unsigned long  BinaryTime [2],
                         FlushDeltaBinTime [2];

   static char  BytesString [32],
                Elapsed [32],
                FormatUser [256],
                StatusString [16],
                String [1024],
                TimeString [32],
                VmsTimeString [32];
   static $DESCRIPTOR (BytesStringDsc, BytesString);
   static $DESCRIPTOR (ElapsedDsc, Elapsed);
   static $DESCRIPTOR (StatusStringDsc, StatusString);
   static $DESCRIPTOR (StringDsc, String);
   static $DESCRIPTOR (TimeStringDsc, TimeString);
   static $DESCRIPTOR (VmsTimeStringDsc, VmsTimeString);

   static struct FAB  LogFileFab;
   static struct RAB  LogFileRab;

   struct {
      short int  buf_len;
      short int  item;
      unsigned char   *buf_addr;
      unsigned short  *ret_len;
   } LnmItems [] =
   {
      { sizeof(LoggingFileName)-1, LNM$_STRING, LoggingFileName, &Length },
      { 0,0,0,0 }
   };

   register unsigned long  *vecptr;

   boolean  PrevLoggingEnabled;
   int  status,
        SetimrStatus,
        SetPrvStatus;
   unsigned long  FaoVector [MAX_FAO_VECTOR];
   char  *FunctionNamePtr;
   char  MiscChars [MAX_FAO_VECTOR * 2];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "Logging() Function: %d\n", Function);

   PrevLoggingEnabled = LoggingEnabled;

   if (Function == LOGGING_OPEN || Function == LOGGING_BEGIN)
   {
      if (Function == LOGGING_OPEN && LoggingEnabled) return (STS$K_ERROR);

      LoggingEnabled = true;

      UserFormatProblem = false;
      if (LoggingFormatUser[0])
         strcpy (FormatUser, LoggingFormatUser);
      else
      if (Config.LogFormat[0])
         strcpy (FormatUser, Config.LogFormat);
      else
         FormatUser[0] = '\0';

      if (strsame (FormatUser, "COMMON", -1))
      {
         LoggingFormatCommonServer = LoggingFormatCombined = false;
         LoggingFormatCommon = true;
         FormatUser[0] = '\0';
      }
      else
      if (strsame (FormatUser, "COMMON_SERVER", -1))
      {
         LoggingFormatCommon = LoggingFormatCombined = false;
         LoggingFormatCommonServer = true;
         FormatUser[0] = '\0';
      }
      else
      if (strsame (FormatUser, "COMBINED", -1))
      {
         LoggingFormatCommon = LoggingFormatCommonServer = false;
         LoggingFormatCombined = true;
         FormatUser[0] = '\0';
      }
      else
      if (FormatUser[0])
      {
         LoggingFormatCommon = LoggingFormatCommonServer =
            LoggingFormatCombined = false;
      }
      else
         LoggingFormatCommon = true;

      if (strsame (LoggingFileName, "SYS$OUTPUT", -1))
         LoggingSysOutput = true;
      else
         LoggingSysOutput = false;
   }

   if (!LoggingEnabled) return (STS$K_ERROR);

   if (Function == LOGGING_FLUSH && !RecordCount) return (SS$_NORMAL);

   if (Function == LOGGING_FLUSH)
   {
      /******************/
      /* flush log file */
      /******************/

      if (!LoggingEnabled) return (STS$K_ERROR);
      if (LoggingSysOutput) return (SS$_NORMAL);

      sys$close (&LogFileFab, 0, 0);
      LoggingFileOpen = false;
      RecordCount = 0;
      return (SS$_NORMAL);
   }

   if (!LoggingFileOpen && !LoggingSysOutput)
   {
      /*********************/
      /* open the log file */
      /*********************/

      if (Debug) fprintf (stdout, "LoggingFileName |%s|\n", LoggingFileName);

      if (!LoggingFileName[0])
      {
         if (VMSnok (status =
             sys$trnlnm (0, &LnmFileDevDsc,
                         &LoggingFileNameLogicalDsc, 0, &LnmItems)))
         {
            LoggingEnabled = PrevLoggingEnabled;
            return (status);
         }
         LoggingFileName[Length] = '\0';
         if (Debug) fprintf (stdout, "LoggingFileName |%s|\n", LoggingFileName);
      }

      if (!LoggingFileNameLength)
         LoggingFileNameLength = strlen(LoggingFileName);

      LogFileFab = cc$rms_fab;
      LogFileFab.fab$l_fna = LoggingFileName;  
      LogFileFab.fab$b_fns = LoggingFileNameLength;
      LogFileFab.fab$l_fop = FAB$M_CIF | FAB$M_SQO;
      LogFileFab.fab$b_fac = FAB$M_PUT;
      LogFileFab.fab$b_rfm = FAB$C_STMLF;
      LogFileFab.fab$b_shr = FAB$M_SHRPUT;
      LogFileFab.fab$b_org = FAB$C_SEQ;
      LogFileFab.fab$b_rat = FAB$M_CR;
   
      /* turn on SYSPRV to allow access to the logging file */
      EnableSysPrv ();

      status = sys$create (&LogFileFab, 0, 0);

      /* turn off SYSPRV */
      DisableSysPrv ();

      if (VMSnok (status))
      {
         if (Debug) fprintf (stdout, "sys$create() %%X%08.08X\n", status);
         LoggingEnabled = PrevLoggingEnabled;
         return (status);
      }

      LogFileRab = cc$rms_rab;
      LogFileRab.rab$l_fab = &LogFileFab;
      LogFileRab.rab$b_mbf = 4;
      LogFileRab.rab$l_rop = RAB$M_EOF | RAB$M_WBH;

      if (VMSnok (status = sys$connect (&LogFileRab, 0, 0)))
      {
         if (Debug) fprintf (stdout, "sys$connect() %%X%08.08X\n", status);
         sys$close (&LogFileFab, 0, 0);
         LoggingEnabled = PrevLoggingEnabled;
         return (status);
      }

      LoggingFileOpen = true;
      RecordCount = 0;
   }

   /***********************/
   /* act on the log file */
   /***********************/

   if (rqptr == NULL)
   {
      sys$gettim (&BinaryTime);
      sys$numtim (&NumericTime, &BinaryTime);
   }
   else
      sys$numtim (&NumericTime, &rqptr->BinaryTime);

   status = sys$fao (&LogFileTimeFaoDsc, &Length, &TimeStringDsc,
                     NumericTime[2],
                     MonthName[NumericTime[1]],
                     NumericTime[0],
                     NumericTime[3],
                     NumericTime[4],
                     NumericTime[5],
                     TimeGmtString);
   /** if (Debug) fprintf (stdout, "sys$fao() %%X%08.08X\n", status); **/
   TimeString[Length] = '\0';

   if (Function == LOGGING_BEGIN || Function == LOGGING_OPEN ||
       Function == LOGGING_END || Function == LOGGING_CLOSE)
   {
      if (Function == LOGGING_BEGIN)
         FunctionNamePtr = "BEGIN";
      else
      if (Function == LOGGING_OPEN)
         FunctionNamePtr = "OPEN";
      else
      if (Function == LOGGING_CLOSE)
         FunctionNamePtr = "CLOSE";
      else
      if (Function == LOGGING_END)
         FunctionNamePtr = "END";

      if (LoggingFormatCommon ||
          LoggingFormatCommonServer ||
          LoggingFormatCombined)
      {
         vecptr = FaoVector;
         *vecptr++ = ServerHostName;
         *vecptr++ = TimeString;
         *vecptr++ = FunctionNamePtr;
         *vecptr++ = ExitStatus;
         if (LoggingFormatCommonServer)
            *vecptr++ = " -";
         else
         if (LoggingFormatCombined)
            *vecptr++ = " - -";
         else
            *vecptr++ = "";

         status = sys$faol (&LogFileBogusFaoDsc, &Length, &StringDsc,
                            &FaoVector);
         /* if (Debug) fprintf (stdout, "sys$faol() %%X%08.08X\n", status); */
         if (VMSnok (status) || status == SS$_BUFFEROVF)
            ErrorExitVmsStatus (status, "sys$faol()", FI_LI);
         String[Length] = '\0';

         if (LoggingSysOutput)
            fprintf (stdout, "%s\n", String);
         else
         {
            LogFileRab.rab$l_rbf = String;
            LogFileRab.rab$w_rsz = Length;
            /** if (Debug) fprintf (stdout, "String |%s|\n", String); **/
            if (VMSnok (status = sys$put (&LogFileRab, 0, 0)))
               if (Debug) fprintf (stdout, "sys$put() %%X%08.08X\n", status);
            RecordCount++;
         }
      }

      if (Function == LOGGING_END || Function == LOGGING_CLOSE)
      {
         /******************/
         /* close log file */
         /******************/

         LoggingEnabled = false;
         if (LoggingSysOutput) return (SS$_NORMAL);
         sys$close (&LogFileFab, 0, 0);
         LoggingFileOpen = false;
         RecordCount = 0;
         return (SS$_NORMAL);
      }

      /* BEGIN and OPEN drop through to initialize a flush timer */
   }

   if (Function == LOGGING_ENTRY && rqptr != NULL)
   {
      /*************************/
      /* format request record */
      /*************************/

      if (LoggingFormatCommon ||
          LoggingFormatCommonServer ||
          LoggingFormatCombined)
      {
         /*******************/
         /* common/combined */
         /*******************/

         vecptr = FaoVector;

         *vecptr++ = rqptr->ClientHostName;

         /* if authentication has not been supplied "place-mark" remote user */
         if (rqptr->RemoteUser[0])
            *vecptr++ = rqptr->RemoteUser;
         else
            *vecptr++ = "-";

         *vecptr++ = TimeString;

         if (rqptr->HttpMethodName[0])
            *vecptr++ = rqptr->HttpMethodName;
         else
            *vecptr++ = "-";

         if (rqptr->ResourcePtr == NULL)
            *vecptr++ = "-";
         else
         if (!rqptr->ResourcePtr[0])
            *vecptr++ = "-";
         else
            *vecptr++ = rqptr->ResourcePtr;

         *vecptr++ = rqptr->ResponseStatusCode;

         if (rqptr->BytesTx)
         {
            sys$fao (&BytesFaoDsc, &Length, &BytesStringDsc, rqptr->BytesTx);
            *vecptr++ = BytesString;
         }
         else
            *vecptr++ = "0";

         if (LoggingFormatCommonServer)
         {
            if (rqptr->ServerHostNamePtr == NULL)
               *vecptr++ = "-";
            else
               *vecptr++ = rqptr->ServerHostNamePtr;
         }
         else
         if (LoggingFormatCombined)
         {
            if (rqptr->HttpRefererPtr == NULL)
               *vecptr++ = "-";
            else
            if (!rqptr->HttpRefererPtr[0])
               *vecptr++ = "-";
            else
               *vecptr++ = rqptr->HttpRefererPtr;

            if (rqptr->HttpUserAgentPtr == NULL)
               *vecptr++ = "-";
            else
            if (!rqptr->HttpUserAgentPtr[0])
               *vecptr++ = "-";
            else
               *vecptr++ = rqptr->HttpUserAgentPtr;
         }
      }
      else
      if (FormatUser[0] || Config.LogFormat[0])
      {
         /***********************/
         /* user-defined format */
         /***********************/

         register int  cnt;
         register char  *cptr, *sptr;
         char  *AbsentPtr;

         if (UserFormatProblem) return (SS$_NORMAL);

         cnt = 0;
         sptr = MiscChars;
         AbsentPtr = "";
         if (FormatUser[0])
            cptr = FormatUser;
         else
            cptr = Config.LogFormat;

         vecptr = FaoVector;

         while (*cptr)
         {
            if (Debug) fprintf (stdout, "|%s|\n", cptr);

            if (cnt > MAX_FAO_VECTOR)
               ErrorExitVmsStatus (0, ErrorLoggingFormatLength, FI_LI);

            if (cnt == 1) AbsentPtr = FaoVector[0];

            switch (*cptr)
            {
               case '!' :
               {
                  cptr++;
                  switch (*(unsigned short*)cptr)
                  {
                     case 'AR' :
                        if (rqptr->AuthRealmPtr == NULL)
                           *vecptr++ = AbsentPtr;
                        else
                        if (!rqptr->AuthRealmPtr[0])
                           *vecptr++ = AbsentPtr;
                        else
                           *vecptr++ = rqptr->AuthRealmPtr;
                        cnt++;
                        cptr += 2;
                        continue;
                     case 'AU' :
                        if (!rqptr->RemoteUser[0])
                           *vecptr++ = AbsentPtr;
                        else
                           *vecptr++ = rqptr->RemoteUser;
                        cnt++;
                        cptr += 2;
                        continue;
                     case 'BB' :
                        if (rqptr->BytesTx)
                        {
                           sys$fao (&BytesFaoDsc, 0, &BytesStringDsc,
                              rqptr->BytesTx - rqptr->ResponseHeaderLength);
                           *vecptr++ = BytesString;
                        }
                        else
                           *vecptr++ = "0";
                        cnt++;
                        cptr += 2;
                        continue;
                     case 'BY' :
                        if (rqptr->BytesTx)
                        {
                           sys$fao (&BytesFaoDsc, 0, &BytesStringDsc,
                                    rqptr->BytesTx);
                           *vecptr++ = BytesString;
                        }
                        else
                           *vecptr++ = "0";
                        cnt++;
                        cptr += 2;
                        continue;
                     case 'CN' :
                        *vecptr++ = rqptr->ClientHostName;
                        cnt++;
                        cptr += 2;
                        continue;
                     case 'CA' :
                        *vecptr++ = rqptr->ClientInternetAddress;
                        cnt++;
                        cptr += 2;
                        continue;
                     case 'EM' :
                        /* durations are measured in microseconds */
                        sys$fao (&ElapsedResponseDurationDsc, 0, &ElapsedDsc,
                                 rqptr->ResponseDuration / 1000);
                        *vecptr++ = Elapsed;
                        cnt++;
                        cptr += 2;
                        continue;
                     case 'ES' :
                        /* durations are measured in microseconds */
                        sys$fao (&ElapsedSecondsDsc, 0, &ElapsedDsc,
                                 rqptr->ResponseDuration / 1000000,
                                 (rqptr->ResponseDuration / 1000) % 1000);
                        *vecptr++ = Elapsed;
                        cnt++;
                        cptr += 2;
                        continue;
                     case 'ME' :
                        *vecptr++ = rqptr->HttpMethodName;
                        cnt++;
                        cptr += 2;
                        continue;
                     case 'PA' :
                        if (rqptr->PathInfoPtr == NULL)
                           *vecptr++ = AbsentPtr;
                        else
                        if (!rqptr->PathInfoPtr[0])
                           *vecptr++ = AbsentPtr;
                        else
                           *vecptr++ = rqptr->PathInfoPtr;
                        cnt++;
                        cptr += 2;
                        continue;
                     case 'QS' :
                        if (rqptr->QueryStringPtr == NULL)
                           *vecptr++ = AbsentPtr;
                        else
                        if (!rqptr->QueryStringPtr[0])
                           *vecptr++ = AbsentPtr;
                        else
                           *vecptr++ = rqptr->QueryStringPtr;
                        cnt++;
                        cptr += 2;
                        continue;
                     case 'RF' :
                        if (rqptr->HttpRefererPtr == NULL)
                           *vecptr++ = AbsentPtr;
                        else
                        if (!rqptr->HttpRefererPtr[0])
                           *vecptr++ = AbsentPtr;
                        else
                           *vecptr++ = rqptr->HttpRefererPtr;
                        cnt++;
                        cptr += 2;
                        continue;
                     case 'RQ' :
                        if (rqptr->ResourcePtr == NULL)
                           *vecptr++ = AbsentPtr;
                        else
                        if (!rqptr->ResourcePtr[0])
                           *vecptr++ = AbsentPtr;
                        else
                           *vecptr++ = rqptr->ResourcePtr;
                        cnt++;
                        cptr += 2;
                        continue;
                     case 'RS' :
                        sys$fao (&StatusFaoDsc, 0, &StatusStringDsc,
                                 rqptr->ResponseStatusCode);
                        *vecptr++ = StatusString;
                        cnt++;
                        cptr += 2;
                        continue;
                     case 'SC' :
                        if (!rqptr->ScriptName[0])
                           *vecptr++ = AbsentPtr;
                        else
                           *vecptr++ = rqptr->ScriptName;
                        cnt++;
                        cptr += 2;
                        continue;
                     case 'SN' :
                        if (rqptr->ServerHostNamePtr == NULL)
                           *vecptr++ = AbsentPtr;
                        else
                           *vecptr++ = rqptr->ServerHostNamePtr;
                        cnt++;
                        cptr += 2;
                        continue;
                     case 'SP' :
                        if (rqptr->ServerPortStringPtr == NULL)
                           *vecptr++ = AbsentPtr;
                        else
                           *vecptr++ = rqptr->ServerPortStringPtr;
                        cnt++;
                        cptr += 2;
                        continue;
                     case 'TC' :
                        *vecptr++ = TimeString;
                        cnt++;
                        cptr += 2;
                        continue;
                     case 'TG' :
                        *vecptr++ = rqptr->GmDateTime;
                        cnt++;
                        cptr += 2;
                        continue;
                     case 'TV' :
                        sys$fao (&VmsTimeFaoDsc, 0, &VmsTimeStringDsc,
                                 &rqptr->BinaryTime);
                        *vecptr++ = VmsTimeString;
                        cnt++;
                        cptr += 2;
                        continue;
                     case 'UA' :
                        if (rqptr->HttpUserAgentPtr == NULL)
                           *vecptr++ = AbsentPtr;
                        else
                        if (!rqptr->HttpUserAgentPtr[0])
                           *vecptr++ = AbsentPtr;
                        else
                           *vecptr++ = rqptr->HttpUserAgentPtr;
                        cnt++;
                        cptr += 2;
                        continue;
                     default :
                        UserFormatProblem = true;
                        continue;
                  }
               }

               case '\\' :
               {
                  cptr++;
                  switch (*cptr)
                  {
                     case '0' :
                        *vecptr++ = "";
                        cnt++;
                        cptr++;
                        continue;
                     case 'n' :
                        *vecptr++ = "\n";
                        cnt++;
                        cptr++;
                        continue;
                     case 'q' :
                        *vecptr++ = "\"";
                        cnt++;
                        cptr++;
                        continue;
                     case 't' :
                        *vecptr++ = "\t";
                        cnt++;
                        cptr++;
                        continue;
                     case '!' :
                        *vecptr++ = "!";
                        cnt++;
                        cptr++;
                        continue;
                     case '\\' :
                        *vecptr++ = "\\";
                        cnt++;
                        cptr++;
                        continue;
                     default :
                        UserFormatProblem = true;
                        continue;
                  }
               }

               default :
               {
                  *sptr = *cptr++;
                  *vecptr++ = sptr++;
                  *sptr++ = '\0';
                  cnt++;
                  continue;
               }
            }
         }

         if (cnt <= 1) UserFormatProblem = true;

         if (UserFormatProblem)
         {
            fprintf (stdout, "%%%s-W-LOG, user format problem\n \\%s\\\n",
                     Utility, FormatUser);
            return (SS$_NORMAL);
         }

         /** if (Debug) fprintf (stdout, "cnt-1: %d\n", cnt-1); **/
         /* now set [0] to the repeat count */
         FaoVector[0] = cnt - 1;
      }
      else
         ErrorExitVmsStatus (0, ErrorSanityCheck, FI_LI);

      /*************/
      /* sys$fao() */
      /*************/

      if (LoggingFormatCommon)
         status = sys$faol (&LogFileCommonFaoDsc, &Length, &StringDsc,
                            &FaoVector);
      else
      if (LoggingFormatCombined)
         status = sys$faol (&LogFileCombinedFaoDsc, &Length, &StringDsc,
                            &FaoVector);
      else
      if (LoggingFormatCommonServer)
         status = sys$faol (&LogFileCommonServerFaoDsc, &Length, &StringDsc,
                            &FaoVector);
      else
      if (FormatUser[0])
         status = sys$faol (&LogFileUserFaoDsc, &Length, &StringDsc,
                            &FaoVector);
      else
         ErrorExitVmsStatus (0, ErrorSanityCheck, FI_LI);

      /** if (Debug) fprintf (stdout, "sys$faol() %%X%08.08X\n", status); **/

      if (VMSnok (status) || status == SS$_BUFFEROVF)
         ErrorExitVmsStatus (status, "sys$faol()", FI_LI);
      String[Length] = '\0';

      /********************/
      /* write the record */
      /********************/

      if (LoggingSysOutput)
         fprintf (stdout, "%s\n", String);
      else
      {
         LogFileRab.rab$l_rbf = String;
         LogFileRab.rab$w_rsz = Length;
         /** if (Debug) fprintf (stdout, "String |%s|\n", String); **/
         if (VMSnok (status = sys$put (&LogFileRab, 0, 0)))
            if (Debug) fprintf (stdout, "sys$put() %%X%08.08X\n", status);
         RecordCount++;
      }
   }

   /*******************************/
   /* flush log file after period */
   /*******************************/

   if (!FlushDeltaBinTime[0])
   {
      if (VMSnok (status = sys$bintim (&FlushDeltaTimeDsc, &FlushDeltaBinTime)))
         ErrorExitVmsStatus (status, "sys$bintim()", FI_LI);
   }
   if (!LoggingFlushTimerOutstanding)
   {
      if (VMSnok (SetimrStatus =
          sys$setimr (0, &FlushDeltaBinTime, &LoggingFlushFileAST,
                      &LoggingFlushFileAST, 0)))
         ErrorExitVmsStatus (SetimrStatus, "sys$setimr()", FI_LI);
      LoggingFlushTimerOutstanding = true;
   }

   return (status);
}

/*****************************************************************************/
/*
Timer AST function to "flush" (close) the log file.
*/

LoggingFlushFileAST ()

{
   Logging (NULL, LOGGING_FLUSH);
   LoggingFlushTimerOutstanding = false;
}

/*****************************************************************************/

